/*
 * Copyright 2014 Open Networking Laboratory
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Application developed by Aitor Zabala Orive
 */
package org.ctpd.olt;

import java.nio.ByteBuffer;
import com.google.common.collect.Maps;
import org.apache.felix.scr.annotations.Activate;
import org.apache.felix.scr.annotations.Component;
import org.apache.felix.scr.annotations.Deactivate;
import org.apache.felix.scr.annotations.Modified;
import org.apache.felix.scr.annotations.Property;
import org.apache.felix.scr.annotations.Reference;
import org.apache.felix.scr.annotations.ReferenceCardinality;
import org.apache.felix.scr.annotations.Service;
import org.onlab.packet.EthType;
import org.onlab.packet.Ethernet;
import org.onlab.packet.VlanId;
import org.onlab.packet.IpPrefix;
import org.onlab.packet.IPv4;
import org.onlab.packet.Ip4Prefix;
import org.onlab.packet.Ip4Address;
import org.onlab.packet.IPv6;
import org.onlab.packet.Ip6Prefix;
import org.onlab.packet.Ip6Address;
import org.onlab.packet.ICMP6;
import org.onlab.packet.MacAddress;
import org.onlab.packet.IpAddress;
import org.onlab.util.Tools;
import org.onlab.util.KryoNamespace;

import org.onlab.packet.TpPort;
import org.onosproject.cfg.ComponentConfigService;
import org.onosproject.core.ApplicationId;
import org.onosproject.core.CoreService;
import org.onosproject.event.AbstractListenerManager;
import org.onosproject.event.Event;
import org.onosproject.incubator.elasticcfg.ElasticConfigService;
import org.onosproject.mastership.MastershipService;
import org.onosproject.net.Host;
import org.onosproject.net.ConnectPoint;
import org.onosproject.net.DeviceId;
import org.onosproject.net.Port;
import org.onosproject.net.PortNumber;
import org.onosproject.net.HostId;
import org.onosproject.net.config.SubjectFactory;
import org.onosproject.net.config.NetworkConfigService;
import org.onosproject.net.config.ConfigFactory;
import org.onosproject.net.config.NetworkConfigEvent;
import org.onosproject.net.config.NetworkConfigListener;
import org.onosproject.net.config.NetworkConfigRegistry;
import org.onosproject.net.config.basics.SubjectFactories;
import org.onosproject.net.device.DeviceEvent;
import org.onosproject.net.device.DeviceEvent.Type;
import org.onosproject.net.device.DeviceListener;
import org.onosproject.net.device.DeviceService;
import org.onosproject.net.flow.criteria.Criterion;
import org.onosproject.net.flow.criteria.PortCriterion;
import org.onosproject.net.flow.criteria.MetadataCriterion;
import org.onosproject.net.flow.criteria.EthCriterion;
import org.onosproject.net.flow.instructions.Instruction;
import org.onosproject.net.flow.instructions.Instructions.OutputInstruction;
import org.onosproject.net.flow.FlowRule;
import org.onosproject.net.flow.FlowRuleService;
import org.onosproject.net.flow.DefaultTrafficSelector;
import org.onosproject.net.flow.DefaultTrafficTreatment;
import org.onosproject.net.flow.TrafficSelector;
import org.onosproject.net.flow.TrafficTreatment;
import org.onosproject.net.flow.criteria.Criteria;
import org.onosproject.net.flow.FlowRuleProviderRegistry;
import org.onosproject.net.flow.FlowEntry;
import org.onosproject.net.flowobjective.DefaultFilteringObjective;
import org.onosproject.net.flowobjective.DefaultForwardingObjective;
import org.onosproject.net.flowobjective.FilteringObjective;
import org.onosproject.net.flowobjective.FlowObjectiveService;
import org.onosproject.net.flowobjective.ForwardingObjective;
import org.onosproject.net.flowobjective.Objective;
import org.onosproject.net.flowobjective.ObjectiveContext;
import org.onosproject.net.flowobjective.ObjectiveError;
import org.onosproject.net.host.HostService;
import org.onosproject.net.packet.PacketContext;
import org.onosproject.net.packet.PacketPriority;
import org.onosproject.net.packet.PacketProcessor;
import org.onosproject.net.packet.PacketService;
import org.onosproject.net.packet.OutboundPacket;
import org.onosproject.net.packet.DefaultOutboundPacket;
import org.onosproject.net.packet.InboundPacket;
import org.onosproject.net.topology.TopologyListener;
import org.onosproject.net.topology.TopologyService;
import org.onosproject.net.topology.TopologyEvent;
import org.onosproject.net.MastershipRole;
import org.ctpd.olt.api.OltDeviceService;
import org.ctpd.olt.api.OltDevice;
import org.onosproject.store.serializers.KryoNamespaces;
import org.onosproject.store.service.Serializer;
import org.onosproject.store.service.StorageService;
import org.onosproject.store.service.ConsistentMap;
import org.onosproject.store.service.AtomicCounter;
import org.onosproject.store.service.DistributedSet;
import org.onosproject.store.service.Versioned;

import org.osgi.service.component.ComponentContext;
import org.slf4j.Logger;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ObjectNode;

import java.util.Collection;
import java.util.Dictionary;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Optional;
import java.util.Properties;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import javax.net.ssl.HttpsURLConnection;

import static com.google.common.base.Strings.isNullOrEmpty;
import static org.onlab.util.Tools.get;
import static org.onlab.util.Tools.groupedThreads;
import static org.slf4j.LoggerFactory.getLogger;

import static java.lang.Math.toIntExact;


import java.io.DataOutputStream;
import java.io.OutputStreamWriter;
import java.io.OutputStream;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.URL;
import java.io.StringWriter;
import java.io.PrintWriter;

import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import java.security.cert.X509Certificate;

import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;

import java.net.MalformedURLException;
import java.net.URL;
import java.io.IOException;
import javax.net.ssl.HostnameVerifier;


import org.onosproject.net.meter.MeterId;

/**
 * Provisions rules on access devices.
 */
@Service
@Component(immediate = true)
public class Olt
        // extends AbstractListenerManager<AccessDeviceEvent, AccessDeviceListener>
        implements OltDeviceService {


    private static final int SUBSCRIBER_FLOW_PRIORITY = 1000;
    private static final int MAX_NUM_OLT_DEF = 256;
    private static final boolean BACKEND_ACTIVATED = true;
    private static final String BACKEND_BASE_URL= "http://192.168.115.154";
    private static final String BACKEND_PORT_EVNT_URL = "/api/onos/olt/client/port";
    private static final String OLT_DEVICE_ID = "of:0001";
    private static final String DEFAULT_XGSPON_ID = "of:000070b3d552312b";
    private static final boolean DEFAULT_KEEP_DATA = true;
    private static final boolean DEFAULT_USE_METER = false;
    private static final boolean DEFAULT_VOLTHA_16 = false;

    private final Logger log = getLogger("log4j.logger.org.ctpdLogger");

    //private final Logger log = getLogger(getClass());

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    protected FlowObjectiveService flowObjectiveService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    protected MastershipService mastershipService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    protected DeviceService deviceService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    protected HostService hostService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    protected TopologyService topologyService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    protected CoreService coreService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    protected FlowRuleService flowRuleService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    protected NetworkConfigRegistry networkConfig;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    protected ComponentConfigService componentConfigService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    protected StorageService storageService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    protected PacketService packetService;


    @Property(name = "subcriberMatchPriority", intValue = SUBSCRIBER_FLOW_PRIORITY,
            label = "Subscriber flow priority")
    private int subcriberMatchPriority = SUBSCRIBER_FLOW_PRIORITY;

    @Property(name = "MAX_NUM_OLT", intValue = MAX_NUM_OLT_DEF,
            label = "Maximum number of OLT that can be registered to this app")
    private int MAX_NUM_OLT = MAX_NUM_OLT_DEF;

    @Property(name = "backendActivated", boolValue = BACKEND_ACTIVATED,
            label = "Send Port data to Backend")
    private boolean backendActivated = BACKEND_ACTIVATED;

    @Property(name = "backendBaseURL", value = BACKEND_BASE_URL,
            label = "Base url of backend")
    private String backendBaseURL = BACKEND_BASE_URL;

    @Property(name = "backendPortEventUrl", value = BACKEND_PORT_EVNT_URL,
            label = "Backend url for port events")
    private String backendPortEventUrl = BACKEND_PORT_EVNT_URL;

    @Property(name = "oltOpenflowSwitchPrefix", value = OLT_DEVICE_ID, label = "OLT Switch Device Id")
    private String oltOpenflowSwitchPrefix = OLT_DEVICE_ID;

    private String[] oltOpenflowSwitchPrefixes = OLT_DEVICE_ID.split(",");

    @Property(name = "keepData", boolValue = DEFAULT_KEEP_DATA, label = "Keep service data when upgrading app")
    private boolean keepData = DEFAULT_KEEP_DATA;

    @Property(name = "useMeter", boolValue = DEFAULT_USE_METER, label = "Use Meter in flows installed")
    private boolean useMeter = DEFAULT_USE_METER;

    // @Property(name = "voltha16", boolValue = DEFAULT_VOLTHA_16, label = "Metadata in voltha 1.6 format")
    // private boolean voltha16 = DEFAULT_VOLTHA_16;

    @Property(name = "xgsponId", value = DEFAULT_XGSPON_ID, label = "XGSPON deviceId")
	private String xgsponId = DEFAULT_XGSPON_ID;

    private ApplicationId appId;

    private DistributedSet<DeviceId> oltDeviceSet;
    private ConsistentMap<DeviceId, PortNumber> oltOutputPort;
    private ConsistentMap<DeviceId, VlanId> oltVlan;
    private ConsistentMap<DeviceId, Map<PortNumber, VlanId>> oltClientPorts;
    private ConsistentMap<DeviceId, Short> oltId;

    private AtomicCounter oltDeviceIdCounter;
    private DeviceListener deviceListener = new InternalDeviceListener();

    @Activate
    public void activate(ComponentContext context) {
        appId = coreService.registerApplication("org.ctpd.olt");
        componentConfigService.registerProperties(getClass());
        modified(context);

        KryoNamespace.Builder serializer = KryoNamespace.newBuilder()
                .register(KryoNamespaces.API).register(VlanId.class)
                .register(DeviceId.class).register(PortNumber.class)
                .register(MacAddress.class).register(IpAddress.class);

        oltClientPorts = storageService.<DeviceId, Map<PortNumber, VlanId>> consistentMapBuilder()
                .withSerializer(Serializer.using(serializer.build()))
                .withName("OLT-data")
                .withApplicationId(appId)
                //.withPurgeOnUninstall()
                .build();

        oltDeviceSet = storageService.<DeviceId>setBuilder()
                .withName("Device-Id-Set")
                .withApplicationId(appId)
                .withSerializer(Serializer.using(serializer.build()))
                //.withPurgeOnUninstall()
                .build()
                .asDistributedSet();

        oltId = storageService.<DeviceId, Short> consistentMapBuilder()
                .withSerializer(Serializer.using(serializer.build()))
                .withName("OLTId to deviceId mapper")
                .withApplicationId(appId)
                //.withPurgeOnUninstall()
                .build();

        oltDeviceIdCounter = storageService.atomicCounterBuilder()
                .withApplicationId(appId)
                .withName("OLT Counter")
                .build()
                .asAtomicCounter();

        oltOutputPort = storageService.<DeviceId, PortNumber> consistentMapBuilder()
                .withSerializer(Serializer.using(serializer.build()))
                .withName("Device id Output port mapper")
                .withApplicationId(appId)
                //.withPurgeOnUninstall()
                .build();

        oltVlan = storageService.<DeviceId, VlanId> consistentMapBuilder()
                .withSerializer(Serializer.using(serializer.build()))
                .withName("Stag mapper")
                .withApplicationId(appId)
                //.withPurgeOnUninstall()
                .build();

        oltDeviceIdCounter.set(1);

        deviceService.addListener(deviceListener);

        log.debug("OnLife: CTpd OLT started");
    }

    @Deactivate
    public void deactivate() {
        componentConfigService.unregisterProperties(getClass(), false);
        deviceService.removeListener(deviceListener);

        /*  Persistent Data */
        if(!keepData)
        {
            oltDeviceSet.destroy();
            oltId.destroy();
            oltDeviceIdCounter.destroy();
            oltClientPorts.destroy();
            oltOutputPort.destroy();
            oltVlan.destroy();
        }
        log.debug("OnLife: CTpd OLT stopped");
    }

    @Override
    public void purgeOLT() {

        oltClientPorts.clear();

        oltDeviceSet.removeAll(oltDeviceSet);
        oltDeviceSet.clear();
        oltId.clear();
        oltOutputPort.clear();
        oltVlan.clear();
        oltDeviceIdCounter.set(1);

        flowRuleService.removeFlowRulesById(appId);


        /*  Initialize again all variables to default state */
        //generateInitialStateVariables();
    }

    /* CLI Action Command */
    @Override
    public String getOLTSubscribers() {
        String result = "";
        for (Map.Entry<DeviceId, Versioned<Map<PortNumber, VlanId>>> entryDevice : oltClientPorts.entrySet()) {
            result = result + entryDevice.getKey().toString() + "\n";
            for (Map.Entry<PortNumber, VlanId> entry : entryDevice.getValue().value().entrySet()) {
                result = result + String.format("\t%s - %s\n", entry.getKey().toString(), entry.getValue().toString());
            }
        }
        return result;
    }

    @Modified
    public void modified(ComponentContext context) {
        Dictionary<?, ?> properties = context != null ? context.getProperties() : new Properties();

        try {
            String s = get(properties, "subcriberMatchPriority");
            subcriberMatchPriority = isNullOrEmpty(s) ? SUBSCRIBER_FLOW_PRIORITY : Integer.parseInt(s.trim());
        } catch (Exception e) {
            subcriberMatchPriority = SUBSCRIBER_FLOW_PRIORITY;
        }

        try {
            String s = get(properties, "MAX_NUM_OLT");
            MAX_NUM_OLT = isNullOrEmpty(s) ? MAX_NUM_OLT_DEF : Integer.parseInt(s.trim());
        } catch (Exception e) {
            MAX_NUM_OLT = MAX_NUM_OLT_DEF;
        }

        try {
            String s = get(properties, "backendActivated");
            backendActivated = isNullOrEmpty(s) ? BACKEND_ACTIVATED : Boolean.parseBoolean(s.trim());
        } catch (Exception e) {
            backendActivated = BACKEND_ACTIVATED;
        }

        try {
            String s = get(properties, "backendBaseURL");
            backendBaseURL = isNullOrEmpty(s) ? BACKEND_BASE_URL : s.trim();
        } catch (Exception e) {
            backendBaseURL = BACKEND_BASE_URL;
        }

        try {
            String s = get(properties, "backendPortEventUrl");
            backendPortEventUrl = isNullOrEmpty(s) ? BACKEND_PORT_EVNT_URL : s.trim();
        } catch (Exception e) {
            backendPortEventUrl = BACKEND_PORT_EVNT_URL;
        }

		try {
            String s = get(properties, "oltOpenflowSwitchPrefix");
            oltOpenflowSwitchPrefix = isNullOrEmpty(s) ? OLT_DEVICE_ID : s.trim();
            oltOpenflowSwitchPrefixes = oltOpenflowSwitchPrefix.split(",");
        } catch (Exception e) {
            oltOpenflowSwitchPrefix = OLT_DEVICE_ID;
            oltOpenflowSwitchPrefixes = oltOpenflowSwitchPrefix.split(",");
        }

        try {
            String s = get(properties, "keepData");
            keepData = isNullOrEmpty(s) ? DEFAULT_KEEP_DATA : Boolean.parseBoolean(s.trim());
        } catch (Exception e) {
            keepData = DEFAULT_KEEP_DATA;
        }

        try {
            String s = get(properties, "useMeter");
            useMeter = isNullOrEmpty(s) ? DEFAULT_USE_METER : Boolean.parseBoolean(s.trim());
        } catch (Exception e) {
            useMeter = DEFAULT_USE_METER;
        }

        try {
            String s = get(properties, "xgsponId");
            xgsponId = isNullOrEmpty(s) ? DEFAULT_XGSPON_ID : s.trim();
        } catch (Exception e) {
            xgsponId = DEFAULT_XGSPON_ID;
        }

    }

    @Override
    public String addDevice(OltDevice device) {

        // Only if new ...
        if (oltDeviceSet.add(device.deviceId)) {
            if (oltDeviceSet.size() >= (short)MAX_NUM_OLT) {
                return "Device: " + device.deviceId.toString() + ", could not be added, MAX number of OLT's reached.\r\n";
            }

            oltClientPorts.put(device.deviceId, new HashMap<PortNumber, VlanId>());

            // Add uplink port to device
            oltOutputPort.put(device.deviceId, device.uplink);

            // We assign olt-id, starting from 2
            oltId.put(device.deviceId, new Short(getAvailableSTag(device.deviceId).toShort()));

            VlanId vlan=getAvailableSTag(device.deviceId);
            
            oltVlan.put(device.deviceId, vlan);
        }

        VlanId svlan = getUpperVlanId(device.deviceId);
        return String.format("{\"vlan\": %s}\r\n", svlan.toString());
    }

    private boolean checkOltPrefix(DeviceId deviceId) {
		for (String oltPrefix: oltOpenflowSwitchPrefixes) {
			if (deviceId.toString().startsWith(oltPrefix)) {
				return true;
			}
		}
		return false;
	}

    @Override
    public String removeDevice(DeviceId deviceId) {

        // If existing ...
        if (oltDeviceSet.remove(deviceId)) {
            oltOutputPort.remove(deviceId);
            oltId.remove(deviceId);
            oltVlan.remove(deviceId);
            oltClientPorts.remove(deviceId);
            flowRuleService.purgeFlowRules(deviceId);
            return "Device: " + deviceId.toString() + ", was removed successfully.\r\n";
        }
        else {
            return "Device: " + deviceId.toString() +", does not exist.\r\n";
        }
    }

    @Override
    public String addSubscriber(ConnectPoint port,  VlanId inVlan, int bandwidth) {

        DeviceId deviceId = port.deviceId();

        Versioned<Map<PortNumber, VlanId>> oltClientsV = oltClientPorts.get(deviceId);
        Versioned<PortNumber> uplinkPortV = oltOutputPort.get(deviceId);

        if (uplinkPortV == null || oltClientsV == null) {
            return "Device: " + deviceId.toString() + ", not found\r\n";
        }

        PortNumber uplinkSubscriber = uplinkPortV.value();

        Map<PortNumber, VlanId> oltClients = oltClientsV.value();
        VlanId cTag = oltClients.get(port.port());

        // It is a new subscriber...
        if (cTag == null) {
            VlanId lowerVlan = getAvailableCTag(deviceId);
            VlanId upperVlan = getUpperVlanId(deviceId);

            log.debug("[addSubscriber]inVlan : {}", inVlan.toShort());

            // Provision flows for this subscriber
            provisionVlansCtpd(deviceId, uplinkSubscriber, port.port(), inVlan, upperVlan, lowerVlan, bandwidth);

            // We update port with assigned ctag
            oltClients.put(port.port(), lowerVlan);
            oltClientPorts.put(deviceId, oltClients);

            return String.format("{ \"vlanUpper\": %d,\"vlanLower\": %d, \"outPort\": %d }\r\n",
                                upperVlan.toShort(),
                                lowerVlan.toShort(),
                                uplinkSubscriber.toLong());
        }
        // It is an already provisioned subscriber...
        else {

                VlanId upperVlan = getUpperVlanId(deviceId);;
                VlanId lowerVlan = cTag;

                // Just re-provision flows for this subscriber
                provisionVlansCtpd(deviceId, uplinkSubscriber, port.port(), inVlan, upperVlan, lowerVlan, bandwidth);

                return String.format("{ \"vlanUpper\": %d,\"vlanLower\": %d, \"outPort\": %d }\r\n",
                                upperVlan.toShort(),
                                lowerVlan.toShort(),
                                uplinkSubscriber.toLong());

        }
    }

    // Escoge un client id para un nuevo cliente, escogera el siguiente mas alto.
    private VlanId getAvailableCTag(DeviceId deviceId) {

        int firstVlanAvailable = 0;

        log.debug("[getAvailableCTag]");

        ArrayList<Short> clientIdsList = new ArrayList<Short>();

        Map<PortNumber, VlanId> ports = oltClientPorts.get(deviceId).value();

        // Sacamos todas las vlan que se están usando...
        for (Map.Entry<PortNumber, VlanId> entry : ports.entrySet()) {
            clientIdsList.add(entry.getValue().toShort());
            log.debug("[getAvailableCTag] ctag used : {}", entry.getValue().toShort());
        }

        // Las ordenamos...
        ArrayList<Short> orderedVlanIds = clientIdsList.stream()
            .sorted()
            .collect(Collectors.toCollection(ArrayList::new));

        // Si el valor de una vlan no coincide con la posición, es un hueco!
        for(int i=0; i<orderedVlanIds.size(); ++i) {
            log.debug("[getAvailableCTag] ctag ordered : {}", orderedVlanIds.get(i));
            if(orderedVlanIds.get(i) > (i+1))
            {
                firstVlanAvailable = i+1;
                log.debug("[getAvailableCTag] available ctag found : {}", firstVlanAvailable);
                break;
            }
        }

        // Si no hay hueco, tomamos el siguiente
        if (firstVlanAvailable==0)
            firstVlanAvailable = orderedVlanIds.size() + 1;

        return VlanId.vlanId((short) firstVlanAvailable);
    }

    // Escoge un client id para un nuevo cliente, escogera el siguiente mas alto.
    private VlanId getAvailableSTag(DeviceId deviceId) {

        int firstVlanAvailable = 0;

        log.debug("[getAvailableSTag]");

        ArrayList<Short> stagList = new ArrayList<Short>();

        // Sacamos todas las vlan que se están usando...
        for ( Versioned<VlanId> stag : oltVlan.values() )
        {
            stagList.add(stag.value().toShort());
            log.debug("[getAvailableSTag] stag used : {}", stag.value().toShort());
        }

        // Las ordenamos...
        ArrayList<Short> orderedVlanIds = stagList.stream()
            .sorted()
            .collect(Collectors.toCollection(ArrayList::new));

        // Si el valor de una vlan no coincide con la posición, es un hueco!
        for(int i=0; i<orderedVlanIds.size(); ++i) {
            log.debug("[getAvailableSTag] stag ordered : {}", orderedVlanIds.get(i));
            if(orderedVlanIds.get(i) > (i+1))
            {
                firstVlanAvailable = i+1;
                log.debug("[getAvailableSTag] available stag found : {}", firstVlanAvailable);
                break;
            }
        }

        // Si no hay hueco, tomamos el siguiente
        if (firstVlanAvailable==0)
            firstVlanAvailable = orderedVlanIds.size() + 1;

        return VlanId.vlanId((short) firstVlanAvailable);
    }

    private VlanId getUpperVlanId(DeviceId deviceId) {
        short olt = oltId.get(deviceId).value().shortValue();
        short upperVlan = (short)((olt) + 1);
        return VlanId.vlanId(upperVlan);
    }

    @Override
    public String removeSubscriber(DeviceId deviceId, PortNumber inPort) {
        Versioned<Map<PortNumber, VlanId>> oltClientsV = oltClientPorts.get(deviceId);

        PortNumber oltOutput = oltOutputPort.get(deviceId).value();
        Map<PortNumber, VlanId> oltClients = oltClientsV.value();


        if (oltClients == null) {
            return "[Error] Device: " + deviceId.toString() + ", does not exist\r\n";
        }

        VlanId cTag = oltClients.get(inPort);

        if (cTag == null) {
            return "[Error] There is no subcriber for this port\n";
        }

        //  Get flow rules installed by this app
        Iterable<FlowRule> flowRules = flowRuleService.getFlowRulesById(appId);
        for (FlowRule flowRule : flowRules) {
            if (flowRule.deviceId().equals(deviceId)) {
                TrafficSelector selector = flowRule.selector();
                TrafficTreatment treatment = flowRule.treatment();
                PortCriterion portCriterion = (PortCriterion)selector.getCriterion(Criterion.Type.IN_PORT);
                OutputInstruction outputInstruction = null;
                MetadataCriterion metadataCriterion = (MetadataCriterion) selector.getCriterion(Criterion.Type.METADATA);

                for (Instruction instruction : treatment.allInstructions()){
                    if (instruction.getClass() == OutputInstruction.class){
                        outputInstruction = (OutputInstruction)instruction;
                        break;
                    }
                }
                if (flowRule.priority() == subcriberMatchPriority) {
                    if (portCriterion.port().equals(inPort) || (outputInstruction != null && outputInstruction.port().equals(inPort)) ) {
                        flowRuleService.removeFlowRules(flowRule);
                    }
                    if (portCriterion.port().equals(oltOutput) ) {
                        if (metadataCriterion!=null)
                        {
                            long vlanFromMetadata = ((metadataCriterion.metadata() >> 32) & 0xfff);
                            if (cTag.equals(VlanId.vlanId((short)vlanFromMetadata)))
                                flowRuleService.removeFlowRules(flowRule);
                        }
                    }
                }
            }
        }

        oltClients.remove(inPort);
        oltClientPorts.put(deviceId, oltClients);
        return "[OK] Success\r\n";
    }

    private void provisionVlansCtpd(DeviceId deviceId, PortNumber uplinkPort,
                                    PortNumber subscriberPort, VlanId subscriberVlan,
                                    VlanId upperVlan, VlanId lowerVlan, int bandwidth) {

        CompletableFuture<ObjectiveError> downFuture = new CompletableFuture();
        CompletableFuture<ObjectiveError> upFuture = new CompletableFuture();

        ForwardingObjective.Builder upFwd = upBuilderCtpd(uplinkPort, subscriberPort, subscriberVlan,
                                                        upperVlan, lowerVlan, bandwidth);

        ForwardingObjective.Builder downFwd = downBuilderCtpd(deviceId, uplinkPort, subscriberPort, subscriberVlan,
                                                            upperVlan, lowerVlan, bandwidth);

        flowObjectiveService.forward(deviceId, upFwd.add(new ObjectiveContext() {
            @Override
            public void onSuccess(Objective objective) {
                upFuture.complete(null);
            }

            @Override
            public void onError(Objective objective, ObjectiveError error) {
                upFuture.complete(error);
            }
        }));

        flowObjectiveService.forward(deviceId, downFwd.add(new ObjectiveContext() {
            @Override
            public void onSuccess(Objective objective) {
                downFuture.complete(null);
            }

            @Override
            public void onError(Objective objective, ObjectiveError error) {
                downFuture.complete(error);
            }
        }));
        //updateInputPortsToFlood(deviceId);

    }

    private ForwardingObjective.Builder downBuilderCtpd(DeviceId deviceId,
                                                    PortNumber uplinkPort,
                                                    PortNumber subscriberPort,
                                                    VlanId subscriberVlan,
                                                    VlanId upperVlan,
                                                    VlanId lowerVlan,
                                                    int bandwidth) {

        TrafficSelector.Builder downstream = DefaultTrafficSelector.builder();

        downstream //.matchEthType(Ethernet.TYPE_IPV6)
                .matchInnerVlanId(lowerVlan)
                .matchVlanId(upperVlan)
                .matchInPort(uplinkPort);

        TrafficTreatment.Builder downstreamTreatment = DefaultTrafficTreatment.builder();

        long metadata;

        if (deviceId.toString().toLowerCase().startsWith(xgsponId))
            metadata = ((((long)bandwidth) << (12 + 32)) | (((long)lowerVlan.toShort()) << 32) | subscriberPort.toLong());
        else
            metadata = subscriberPort.toLong();

        if(useMeter)
            downstream.matchMetadata(metadata);

        downstreamTreatment.popVlan()
                        .setVlanId(subscriberVlan)
                        .setOutput(subscriberPort);

        return DefaultForwardingObjective.builder()
                .withFlag(ForwardingObjective.Flag.VERSATILE)
                .withPriority(subcriberMatchPriority)
                .makePermanent()
                .withSelector(downstream.build())
                .fromApp(appId)
                .withTreatment(downstreamTreatment.build());
    }

    private ForwardingObjective.Builder upBuilderCtpd(PortNumber uplinkPort,
                                                PortNumber subscriberPort,
                                                VlanId subscriberVlan,
                                                VlanId upperVlan,
                                                VlanId lowerVlan,
                                                int bandwidth) {

        TrafficSelector.Builder upstream = DefaultTrafficSelector.builder();
        upstream .matchVlanId(subscriberVlan) //.matchEthType(Ethernet.TYPE_IPV6)
                .matchInPort(subscriberPort);

        TrafficTreatment.Builder upstreamTreatment = DefaultTrafficTreatment.builder();

        long metadata = bandwidth;

        if(useMeter)
            upstream.matchMetadata(metadata);

        upstreamTreatment.pushVlan()
                        .setVlanId(lowerVlan)
                        .pushVlan()
                        .setVlanId(upperVlan)
                        .setOutput(uplinkPort);

        return DefaultForwardingObjective.builder()
                .withFlag(ForwardingObjective.Flag.VERSATILE)
                .withPriority(subcriberMatchPriority)
                .makePermanent()
                .withSelector(upstream.build())
                .fromApp(appId)
                .withTreatment(upstreamTreatment.build());
    }

    private void sendPortEventToBackend(DeviceId deviceId, String status, Port port) throws Exception {

        if (backendBaseURL.contains("https")){
            log.debug("Using HTTPS to connect to backend");
            sendHttpsPortEventToBackend(deviceId, status, port);
        }else if(backendBaseURL.contains("http")){
            log.debug("Using HTTP to connect to backend");
            sendHttpPortEventToBackend(deviceId, status, port);

        }
    }

    private void sendHttpPortEventToBackend(DeviceId deviceId, String status, Port port) throws Exception {

        PortNumber portNumber = port.number();

        String url = backendBaseURL + backendPortEventUrl;
        log.debug(url);
        URL obj = new URL(url);
        HttpURLConnection connection = (HttpURLConnection) obj.openConnection();


        connection.setRequestMethod("POST");
        connection.setRequestProperty("content-type", "application/json");
        connection.setRequestProperty("Accept", "application/json");

        // Send post request
        connection.setDoOutput(true);
        //connection.connect();

        //DataOutputStream wr = new DataOutputStream(connection.getOutputStream());
        OutputStream os = connection.getOutputStream();
        OutputStreamWriter wr = new OutputStreamWriter(os, "UTF-8");

        ObjectMapper objectMapper = new ObjectMapper();
        ObjectNode objNode = objectMapper.createObjectNode();

        //Get port name and port mac to construct serial id

        String portName = port.annotations().value("portName");
        String portMac = port.annotations().value("portMac");
        String speed = null;

        if(portName.contains("1G_")) {
            speed = "1000";
            portName = portName.replace("1G_","");
        } else if(portName.contains("10G_")) {
            speed = "10000";
            portName = portName.replace("10G_","");
        } else {
            speed = "1000";
        }

        //log.info("[sendPortEventToBackend] Print port name and port mac: {} {}", portName, portMac);
        String serialId = portName.concat(portMac);

        log.debug("[sendPortEventToBackend] Getting serial Id from port parameters: {}", serialId);

        objNode.put("port", portNumber.toString());
        objNode.put("olt", deviceId.toString());
        objNode.put("status", status);
        objNode.put("speed", speed);
        objNode.put("serialId",serialId);

        wr.write(objNode.toString());

        wr.flush();
        wr.close();

        int responseCode = connection.getResponseCode();
        log.debug("\nSending 'POST' request to URL : " + url);
        log.debug("Response Code : " + responseCode);

        // Disconnect
        connection.disconnect();
    }

    private void sendHttpsPortEventToBackend(DeviceId deviceId, String status, Port port) throws Exception {

        PortNumber portNumber = port.number();

        String url = backendBaseURL + backendPortEventUrl;
        log.debug(url);
        URL obj = new URL(url);
        SSLContext sc;
        HttpsURLConnection connection;

        TrustManager[] trustAllCerts = new TrustManager[]{new X509TrustManager(){
		    public X509Certificate[] getAcceptedIssuers(){return null;}
		    public void checkClientTrusted(X509Certificate[] certs, String authType){}
		    public void checkServerTrusted(X509Certificate[] certs, String authType){}
        }};

        try {

            sc = SSLContext.getInstance("SSL");
            sc.init(null, trustAllCerts, new SecureRandom());
            SSLContext.setDefault(sc);
            HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());
            connection = (HttpsURLConnection)obj.openConnection();
            connection.setHostnameVerifier(new HostnameVerifier() {
                public boolean verify(String host, SSLSession sess) {
                    return true;
                }
            });

            connection.setRequestMethod("POST");
            connection.setRequestProperty("content-type", "application/json");
            connection.setRequestProperty("Accept", "application/json");

            // Send post request
            connection.setDoOutput(true);
            //connection.connect();

            //DataOutputStream wr = new DataOutputStream(connection.getOutputStream());
            OutputStream os = connection.getOutputStream();
            OutputStreamWriter wr = new OutputStreamWriter(os, "UTF-8");

            ObjectMapper objectMapper = new ObjectMapper();
            ObjectNode objNode = objectMapper.createObjectNode();

            //Get port name and port mac to construct serial id

            String portName = port.annotations().value("portName");
            String portMac = port.annotations().value("portMac");
            String speed = null;

            if(portName.contains("1G_")) {
                speed = "1000";
                portName = portName.replace("1G_","");
            } else if(portName.contains("10G_")) {
                speed = "10000";
                portName = portName.replace("10G_","");
            } else {
                speed = "1000";
            }

            //log.info("[sendPortEventToBackend] Print port name and port mac: {} {}", portName, portMac);
            String serialId = portName.concat(portMac);

            log.debug("[sendPortEventToBackend] Getting serial Id from port parameters: {}", serialId);

            objNode.put("port", portNumber.toString());
            objNode.put("olt", deviceId.toString());
            objNode.put("status", status);
            objNode.put("speed", speed);
            objNode.put("serialId",serialId);

            wr.write(objNode.toString());

            wr.flush();
            wr.close();

            int responseCode = connection.getResponseCode();
            log.debug("\nSending 'POST' request to URL : " + url);
            log.debug("Response Code : " + responseCode);

            // Disconnect
            connection.disconnect();

        } catch (Exception e) {
            log.error("Error connecting with Backend");
        }

    }

    private void processPortEvent(DeviceId deviceId, Port port, DeviceEvent.Type eventType) {
        String portEvent = "UNKNOWN";

        if (eventType == DeviceEvent.Type.PORT_ADDED)
            portEvent = "PORT_UP";
        else if(eventType == DeviceEvent.Type.PORT_REMOVED)
            portEvent = "PORT_DOWN";
        else {
            if (port.isEnabled()) {
                portEvent = "PORT_UP";
            }
            else {
                portEvent = "PORT_DOWN";
            }
        }

        //PortNumber portNumber = port.number();
        //Get port name and port mac to construct serial id

        log.debug("[InternalDeviceListener] Port:{} Status:{}", port.number().toString(), portEvent);
        try {
                Versioned<PortNumber> uplinkPortV = oltOutputPort.get(deviceId);
                if(uplinkPortV!=null)
                    if(!port.number().equals(uplinkPortV.value()))
                        sendPortEventToBackend(deviceId, portEvent, port);
        } catch (Exception e) {
            log.error("[InternalDeviceListener] Error sending to backend", e);
        }
    }


    private class InternalDeviceListener implements DeviceListener {

        @Override
        public void event(DeviceEvent event) {
            log.debug("[InternalDeviceListener] DeviceEvent received");
            if (backendActivated) {
                DeviceId deviceId = event.subject().id();
                MastershipRole role = deviceService.getRole(deviceId);
                if (role == MastershipRole.MASTER) { /* Only send port event when node is Master */
                    // Only if OLT
                    log.debug("[InternalDeviceListener] Event from device {}", deviceId.toString());
                    if (checkOltPrefix(deviceId))
                    {
                        DeviceEvent.Type type = event.type();
                        log.debug("[InternalDeviceListener] DeviceEvent Type:{}", type);
                        if (type == DeviceEvent.Type.PORT_ADDED || type == DeviceEvent.Type.PORT_UPDATED || type == DeviceEvent.Type.PORT_REMOVED) {
                            Port port = event.port();
                            Thread t1 = new Thread(new Runnable() {
                                public void run() {
                                    processPortEvent(deviceId, port, type);
                                }
                            });
                            t1.start();
                        }
                    }
                }
            }
        }
    }
}
