/*
 * Copyright 2014 Open Networking Laboratory
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Application developed by Aitor Zabala Orive
 */
package org.ctpd.olt;

import javax.ws.rs.DELETE;
import javax.ws.rs.POST;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.io.InputStream;
import javax.ws.rs.Consumes;
import java.util.List;
import java.util.Map;
import java.util.LinkedHashMap;

import org.onlab.packet.VlanId;
import org.onlab.packet.Ip4Address;
import org.onlab.packet.IpAddress;
import org.onlab.packet.MacAddress;
import org.onlab.packet.IpPrefix;
import org.onosproject.net.ConnectPoint;
import org.onosproject.net.DeviceId;
import org.onosproject.net.PortNumber;
import org.ctpd.olt.api.OltDeviceService;
import org.ctpd.olt.api.OltDevice;
import org.ctpd.olt.api.OltVlanPort;
import org.onosproject.rest.AbstractWebResource;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.fasterxml.jackson.databind.node.ArrayNode;


/**
 * OLT REST APIs.
 */

@Path("oltapp")
public class OltWebResource extends AbstractWebResource {
    /**
     * Provision a subscriber.
     *
     * @param device device id
     * @param inport input port number
     * @param id vlan id
     * @param bandwidth bandwidth
     * @return 200 OK
     */

    @POST
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    @Path("{device}")
    public Response provisionSubscriber(@PathParam("device")String device, InputStream stream) {
        List<OltVlanPort> ports = null;
        DeviceId deviceId = DeviceId.deviceId(device);
        ObjectMapper mapper = new ObjectMapper();
        try {
            ports = mapper.readValue(stream,
                    mapper.getTypeFactory().constructCollectionType(List.class, OltVlanPort.class));
        } catch (Exception e) {
            return Response.status(400).entity("DESERIALIZING ERROR:\n" + e.toString() + "\n").build();
        }

        OltDeviceService service = get(OltDeviceService.class);
        Map<String, OltVlanPort> data = new LinkedHashMap<String, OltVlanPort>();
        String result = "";
        for (OltVlanPort port : ports) {
            ConnectPoint connectPoint = new ConnectPoint(deviceId, port.getPort());
            String response = service.addSubscriber(connectPoint, port.getVlan(), port.getBandwidth());
            if (response.contains("not found")) {
                return Response.status(404).entity(response + "\n").build();
            }
            result += (response + "\n");
        }

        //mapper.configure(SerializationFeature.INDENT_OUTPUT, true);
        return ok(result).build();
    }

    // /**
    //  * Add new Device.
    //  *
    //  * @param device device id
    //  * @return 200 OK
    //  */
    // @POST
    // @Produces(MediaType.APPLICATION_JSON)
    // @Path("{device}")
    // public Response addDevice( @PathParam("device")String device ) {
    //     OltDeviceService service = get(OltDeviceService.class);
    //     DeviceId deviceId = DeviceId.deviceId(device);

    //     String response = service.addDevice(deviceId);
    //     return ok(response).build();
    // }

    /**
     * Add new Device, using JSON data
     *
     * @param device device id
     * @param uplink_port
     * @return 200 OK
     */
    @POST
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    @Path("register")
    public Response addDevice(InputStream stream) {
        List<OltDevice> devices = null;
        ObjectMapper mapper = new ObjectMapper();
        try {
            devices = mapper.readValue(stream,
                    mapper.getTypeFactory().constructCollectionType(List.class, OltDevice.class));
        } catch (Exception e) {
            return Response.status(400).entity("DESERIALIZING ERROR:\n" + e.toString() + "\n").build();
        }

        OltDeviceService service = get(OltDeviceService.class);
        Map<String, OltDevice> data = new LinkedHashMap<String, OltDevice>();
        String result = "";
        for (OltDevice device : devices) {
            String response = service.addDevice(device);
            data.put(device.deviceId.toString(), device);
            result += (response + "\n");
        }
        return ok(result).build();
    }

    /**
     * Remove Device.
     *
     * @param device device id
     * @return 200 OK
     */

    @DELETE
    @Produces(MediaType.APPLICATION_JSON)
    @Path("{device}")
    public Response removeDevice( @PathParam("device")String device ) {
        OltDeviceService service = get(OltDeviceService.class);
        DeviceId deviceId = DeviceId.deviceId(device);

        String response = service.removeDevice(deviceId);
        return ok(response).build();
    }

    /**
     * Remove subscriber.
     *
     * @param inport input port number
     * @return 200 OK
     */

    @DELETE
    @Produces(MediaType.APPLICATION_JSON)
    @Path("{device}/{inPort}/{inVlan}")
    public Response removeSubscriber(
        @PathParam("device")String device,
        @PathParam("inPort")long inPort,
        @PathParam("inVlan")short inVlan){
        OltDeviceService service = get(OltDeviceService.class);
        DeviceId deviceId;
        deviceId = DeviceId.deviceId(device);

        ObjectMapper mapper = new ObjectMapper();
        ObjectNode objNode = mapper.createObjectNode();

        PortNumber inPortNumber = PortNumber.portNumber(inPort);
        VlanId inVlanId = VlanId.vlanId(inVlan);
        String result = service.removeSubscriber(deviceId, inPortNumber);
        if (result.contains("[OK]") && !result.contains("[Error]")) {
            return ok(result).entity(objNode.put("response", result)).build();
        }
        else
            return Response.status(400).entity(objNode.put("response", result)).build();
    }
}
